//
// Class to determine interval spectrum given LP coefficients
//

#include <LP.h>
#include <complex.h>
#include <stdio.h>

#ifndef SPECTRUM_H
#define SPECTRUM_H

class Spectrum {

 public:
 
  Spectrum();
  Spectrum(LP &, int);
  Spectrum(const double * const, int, int); 		// set gain to unity
  Spectrum(const double * const, double, int, int); 	// set gain manually
  ~Spectrum();
  
  int length(void);			// Get N
  double operator-(Spectrum &);		// Euclidean distance = spectrum1 - spectrum2
  double &operator[](int);
  Spectrum &operator=(Spectrum &);  

 private:
 
  void doit(void);
 
  double * data;	// Spectrum output
  complex<double> * W;		// Twiddle
  
  int N;			// Number of samples
  LP cf;			// LPCs
  double gain;			// Gain, probably just unity
};


int Spectrum::length(void){

 return N;

}


void Spectrum::doit(void){

// Calculate the DFT of the LPCs, because that's the only sense i can make of it
// Adapted from the DFT algorithm in "C Language Algorithms for Digital Signal Processing"
//
// N = spectrum samples (32 according to Lewis)
// 
//

 complex<double> unity(1.0,0.0);

// cout << "Assigning output space\n";
 data = new double[N];

// cout << "Assigning coefficient space\n";
 W = new complex<double>[N];
 
 double arg = 8.0 * atan(1.0)/N;	// 2PI/N
// cout << "Calculating coefficients\n";
 for (int i = 0; i < N; i++){
 
  complex<double> tmp( cos(arg*i), sin(arg*i) );
  W[i] = tmp;
 
 } // i
 
// from now on, exp(-j*2*PI*n*k/N) -> W[n*k % N]
// Perform N-point DFT over cf.order() samples

// cout << "Doing DFT\n";
 for (int k = 0; k < N; k++){

  complex<double> currentK(cf[0],0.0); // exponential = 1 for k = 0

  for (int n = 1; n < cf.order(); n++){

   complex<double> tmp(cf[n],0.0);
   currentK += tmp * W[n*k % N];

  } // n

  data[k] = gain / (1.0 - abs(currentK));

 } // k

 
} // doit


Spectrum::Spectrum(LP &lpc, int Nremote){

// cout << "Making Spectrum object\n";

 cf = lpc;
 N = Nremote;
 W = NULL; data = NULL;
 gain = 1.0; 
 
 // go away and do the spectrum stuff
 doit();

}


Spectrum::Spectrum(){

// Default constructor

 printf("Hello\n");

 N = 0; // use this to test if there is data to work on
 W = NULL; data = NULL;
 gain = 1.0; 

 // don't doit()

}


Spectrum::Spectrum(const double * const a, int p, int Nremote){

 N = Nremote;
 cf = LP(a, p);
 W = NULL; data = NULL;
 gain = 1.0; 

 // Have LPCs & N, go away and work out the spectrum
 doit();

}


Spectrum::Spectrum(const double * const a, double G, int p, int Nremote){

 N = Nremote;
 cf = LP(a, p);
 W = NULL; data = NULL;
 gain = G; 

 // Have LPCs & N, go away and work out the spectrum
 doit();

}


 

Spectrum::~Spectrum(){

 if (data != NULL) delete [] data;
 if (W != NULL) delete [] W;

}


double Spectrum::operator-(Spectrum &op2){

 double distance = -1.0;
 complex<double> *d = new complex<double>[N];
 complex<double> dc(0.0, 0.0);

// One way of doing it

 if (N == op2.N){
 
  for (int i = 0; i < N; i++)
   distance += pow( ( data[i] - op2.data[i] ), 2.0 );

   distance = sqrt(distance); // Don't really have to root it

 }
 
// Another way of doing it ("Computer Speech Processing")
// dj^2 = (t - rj)^t * (t - rj)
// Uses lots of overloaded complex ops
//
// if (N == op2.N){
// 
//  for (int i = 0; i < N; i++) // Do all terms
//   d[i] = pow( ( data[i] - op2.data[i] ), data[i]) *
//               ( data[i] - op2.data[i] );
//               
//   dc += d[i];
//   distance = abs(sqrt(dc));
//
// }
 
 delete [] d;
 return distance; // will be -1.0 if non-equal spectra

}


double &Spectrum::operator[](int i){

 assert(i >= 0 && i < N);
 
 return data[i];

}


Spectrum &Spectrum::operator=(Spectrum &right){


 printf("Freeing existing data\n");

 delete [] data;
 delete [] W;
 
 printf("Assigning static members (%d)\n", right.N);

 N = right.N;
 cf = right.cf;

 if (N > 0){

  printf("Allocating data\n");
  W = new complex<double>[N];
  data = new double[N];

 }

 printf("Copying members\n");

 for (int i = 0; i < N; i++){

  data[i] = right[i];
  W[i] = right.W[i];

 }
 printf("done\n");

 return * this;
 
}



#endif
