/*********************************************************************

Linear: An experimental speech processing utility by Matthew Maude
m.maude@excite.com / u9700143@bournemouth.ac.uk

*********************************************************************/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <sndfile.h>
#include <string.h>
#include <math.h>

#include <Spectrum.h>


#define NO_PRE_EMPHASIS
#undef NO_PRE_EMPHASIS

#define NO_PITCH
#undef NO_PITCH

#define CONFIG_FILE	"linear.config"

#define PITCH_RESET	0
#define PITCH_TRIGGER	1

#define PITCH_TRIG_LEVEL	0.8
#define PITCH_REL_LEVEL	0.7

#define PI		(4.0*atan(1.0))

#define ERROR_LOG	"my.errors"

#define WINDOW_HAMMING	0.54
#define WINDOW_HANNING	0.5
#define WINDOW_HANN	0.5

#define EMPHASIS_MIN	0.960
#define EMPHASIS_MID	0.975
#define EMPHASIS_MAX	0.990
#define EMPHASIS_NONE	0.0

#define MATCH_EUCLIDEAN	1
#define MATCH_SPECTRUM	2
#define MATCH_MULTIPLY	3

#define DFT_SAMPLES	32	// See Lewis

#define K		512

#define I		0.02	/* 20ms, 160 samples @ 8kHz, use L = 2I */
#define Nms		0.04	/* 2 * I */
#define P		0.015	/* Pitch estimate window */
#define C		0.005	/* Pitch estimate window (zero-crossing method) */



/* Function prototypes */


void Window(double * buffer, int N, double a);
double R(double * window, int winLength, int k);
double durbin(double * a, double * window, int winLength, int p);
double parcor(double * window, int winLength, int j, double E, double * a);
double * preEmphasis(double ** samples, int N, double a1);
int euclideanDistance(double * lpc, double * targets, int p, int n, double * distance, int l);
int * match(double * lpc, int p, double * distance, int intervals, double * targets, int n, int type);
void normaliseTargets(double * targets, int n, int p);
int multiply(double * lpc, double * targets, int p, int n);
double absd(double d);
int readTarget(double ** targets, char * filename, double emphasis);
void threshold(double * buffer, int * indices, int samples, int Is, double threshold);
void spectra(int * indices, double * distance, int intervals, double * lpc, double * targets, int p, int n);
double pitchEstimate(double * correlation, int maxK, double triggerThreshold, double resetThreshold);

double * normaliseCovariance_bak(double * covariance, int k);
double * autoCovariance_bak(double * samples, int window, int maxK);

double * autoCovariance(double * samples, int window, int maxK, int offset);
double * normaliseCovariance(double * covariance, int k);

int readLine(FILE * filePtr, char * buffer, int bufsize);
int loadConfig(double ** targets, int * p, double * thresh, double * trigger, double * release, double * emphasis);
int sign(double x);
double * ZeroCrossing(double * samples, int nSamples, int window);

double average(double * data, int n);
double G(double * lpc, int order, double * window, int Ns);



/* Global variables, to keep function headers small */



double * gain = NULL;

int order = -1;
int n_targets = 0;

FILE * errorlog;





/*********************************************************************

Main

*********************************************************************/

int main(int argc, char ** argv){
 
 FILE * out;

 SNDFILE * snd = NULL;
 SF_INFO * snd_info = (SF_INFO *) malloc(sizeof(SF_INFO));

 double max = 0;
 double * buffer = NULL;	/* Raw audio samples */
 double * window = NULL;
 double * lpc = NULL;		/* All LPC's*/
 double * tmp_target = NULL;
 double * targets = NULL;
 double * covariance = NULL;	/* Autocovariance data */
 double * correlation = NULL;	/* Autocorrelation data */
 double * pitch = NULL;		/* Pitch data */
 double * crossings = NULL;
 double * distance = NULL;
 double * error = NULL;

 double tmp_pitch = 0.0;
 double thresh = 0.0;
 double trigger = PITCH_TRIG_LEVEL;
 double release = PITCH_REL_LEVEL;
 double emphasis = EMPHASIS_NONE;

 char filename[256] = "";
 char src_file[256] = "";
 char target_file[256] = "";

 int status = -1;
 int Ps = 0;			/* Pitch esimate window (samples) */
 int Is = 0;			/* Window interval I (samples) */
 int Ns = 0;			/* Window length (2*Is) */
 int Cs = 0;
 int i = 0, j = 0;

 int * indices = NULL;

 
 if (argc < 2){

  printf("Source file: ");
  scanf("%s", src_file);

 } else strcpy(src_file,argv[1]);

 /* Open a log to record errors and stuff */

 sprintf(filename,"%s.errors",src_file);
 printf("Opening error log \"%s\"\n",filename);

 if (!(errorlog = fopen(filename,"w"))){

  printf("Failed to open error log\n");
  exit(0);

 }


 if (loadConfig(&targets, &order, &thresh, &trigger, &release, &emphasis) == -1){

  printf("Predictor order: ");
  scanf("%d", &order);

  while (order <= 0){

   printf("Predictor order (p > 0): ");
   scanf("%d", &order);

  }

  printf("Threshold: ");
  scanf("%lf", &thresh);

  while (thresh < 0.0){

   printf("Threshold (> 0): ");
   scanf("%lf", &thresh);

  }


  printf("Target files (index 0 is neutral / no sound), EOF to end\n");

  printf("Target %d: ", n_targets+1);
  scanf("%s",target_file);

  while (!feof(stdin)){

   /* Process target file */
   status = readTarget(&targets, target_file, emphasis);

   printf("Target %d: ", n_targets+1);
   scanf("%s",target_file);

  }


 } // if loadConfig

 printf("\nRead %d targets\n",n_targets);

 printf("Processing...\n");

 snd = sf_open_read(src_file,snd_info);

 /* Check for known file type here */

 status = sf_format_check(snd_info);
 printf("Format Check:\t%d\n",status);

 printf("Sample Rate:\t%d\nSamples:\t%d\nChannels:\t%d\nFormat:\t%d\n",
  snd_info->samplerate,snd_info->samples,snd_info->channels,snd_info->format);

 max = sf_signal_max(snd);
 printf("Max Sample:\t%f\n",max);

 if (snd_info->format & SF_FORMAT_WAV == SF_FORMAT_WAV){

  printf("WAV file detected\n");

 }

 buffer = (double *) malloc(sizeof(double)*snd_info->channels*snd_info->samples);

 status = sf_readf_double(snd,buffer,snd_info->samples,1);
 printf("Frames read:\t%d\n",status);


 printf("Doing stuff\n");



/* PRE-EMPHASIS */

#ifndef NO_PRE_EMPHASIS
  preEmphasis(&buffer, snd_info->samples, emphasis);
#endif



  Is = (int) ((double) snd_info->samplerate * I );	/* Length of window interval (samples) */
  Ns = (int) ((double) snd_info->samplerate * Nms );	/* Length of window (samples) */
  Ps = (int) ((double) snd_info->samplerate * P );	/* Length of window (samples) */

/* AUTOCORRELATION & PITCH ESTIMATION */

#ifndef NO_PITCH

  printf("Calculating pitch estimate (%d, %d, %.3f, %.3f) ", Ps, K, trigger, release);
  fflush(stdout);

  pitch = (double *) malloc(sizeof(double)*snd_info->channels*snd_info->samples);

  for (i = 0; i < snd_info->samples; i++){

   if (i < snd_info->samples - K - Ps){

   covariance = autoCovariance(buffer,Ps,K,i);
   correlation = normaliseCovariance(covariance,K);

   if (i == 2000){ //

    printf("Writing Covariance sample\n");

    sprintf(filename,"tmp.cov");
    out = fopen(filename,"w");
    for (j = 0; j < K; j++) fprintf(out,"%f\n",covariance[j]);
    fclose(out);

    sprintf(filename,"tmp.cor");
    out = fopen(filename,"w");
    for (j = 0; j < K; j++) fprintf(out,"%f\n",correlation[j]);
    fclose(out);

   }

   tmp_pitch = pitchEstimate(correlation, K, 0.8,0.7);

   if (tmp_pitch == -1) /* Failed to compute pitch */ {

    fprintf(errorlog,"Failed to compute pitch at sample %d\n",i);

   /* Hold previous pitch */

    if (i > 0) pitch[i] = pitch[i-1]; else pitch[i] = 0.0; /* */

   } else pitch[i] = (double) snd_info->samplerate / tmp_pitch;


   if (covariance) free(covariance);
   if (correlation) free(correlation);

   } else pitch[i] = 0.0;

   if (i % Is == Is - 1){

    printf(".");
    fflush(stdout);

   }

  } /* i */

#endif


 printf(" done\n");

 free(window);





/* Pitch estmation by zero-crossing count */

  Cs = (int) ((double) snd_info->samplerate * C );	/* Length of window (samples) */
  crossings = ZeroCrossing(buffer,snd_info->samples*snd_info->channels,Cs);





/* LINEAR PREDICTION */


  printf("Calculating LPCs (Is: %d, Ns: %d) ", Is, Ns);

  lpc = (double *) malloc(sizeof(double) * (order+1) * snd_info->samples / Is);
  error = (double *) malloc(sizeof(double) * snd_info->samples / Is);
  gain = (double *) malloc(sizeof(double) * snd_info->samples / Is);
  window = (double *) malloc(sizeof(double) * Ns);

  for (i = 0; i < snd_info->samples / Is; i++){

   /* if (i > 0) lseek(stdout,-1,SEEK_CUR); */
   printf(".");

   fflush(stdout);
   memcpy(window, buffer+(i*Is), Ns);
   Window(window, Ns, WINDOW_HAMMING);
   error[i] = durbin(lpc+i*(order+1), window, Ns, order);
   gain[i] = G(lpc+i*(order+1), order, window, Ns);

   /* NOTE: a[0] is unused */

  }

  printf(" done\n");

  free(window);


/* Pattern Matching */

  distance = (double *) malloc(sizeof(double) * snd_info->samples / Is);
  if (n_targets > 0){
   indices = match(lpc, order, distance, snd_info->samples/Is, targets, n_targets, MATCH_SPECTRUM);
   threshold(buffer, indices, snd_info->samples, Is, thresh);
  }


/* OUTPUT */

 if (n_targets > 0){
 sprintf(filename,"%s.verbose",src_file);
 printf("Writing stuff (%s)\n",filename);

 if (out = fopen(filename,"w")){

   for (i = 0; i < snd_info->samples / Is; i++){

    fprintf(out,"%d gain %f err %f target %d (%f)\n",
     i,gain[i],error[i],indices[i],distance[i]);

   } /* i */

   fprintf(out,"\n\n");

  fclose(out);

 } else printf("Failed to open output file\n");
 }

 if (n_targets > 0){
 sprintf(filename,"%s.targets.chan",src_file);
 printf("Writing blend target indices (%s)\n",filename);

 if (out = fopen(filename,"w")){

   for (i = 0; i < snd_info->samples / Is; i++){

    fprintf(out,"%d\n",indices[i]);

   } /* i */

   fprintf(out,"\n\n");

  fclose(out);

 } else printf("Failed to open output file\n");
 }


 sprintf(filename,"%s.lpc.chan",src_file);
 printf("Writing LPCs (%s)\n",filename);

 if (out = fopen(filename,"w")){

  for (i = 0; i < snd_info->samples / Is; i++){

   for (j = 0; j < order; j++){

    fprintf(out,"%f\t",lpc[i*(order+1)+j+1]);

   } /* j */

   fprintf(out,"\n");

  } /* i */

  fclose(out);

 } else printf("Failed to open output file\n");


 sprintf(filename,"%s.lpc",src_file);
 printf("Writing LPCs (%s)\n",filename);

 if (out = fopen(filename,"w")){

  for (j = 0; j < ORDER; j++){

   for (i = 0; i < snd_info->samples / Is; i++){

    fprintf(out,"%d\t%f\n",i*Is,lpc[i*(ORDER+1)+j+1]);

   } /* i */

   fprintf(out,"\n\n");

  } /* j */

  fclose(out);

 } else printf("Failed to open output file\n");

#ifndef NO_PITCH

 sprintf(filename,"%s.pitch.chan",src_file);
 printf("Writing pitch data (%s)\n",filename);

 if (out = fopen(filename,"w")){

  for (i = 0; i < snd_info->samples - (Ps + K); i++){

   fprintf(out,"%f\n",pitch[i]);

  }

  fclose(out);

 } else printf("Failed to open output file\n");


#endif

 sprintf(filename,"%s.crossings.chan",src_file);
 printf("Writing zero-crossing data (%s)\n",filename);

 if (out = fopen(filename,"w")){

  for (i = 0; i < snd_info->samples; i++){

   fprintf(out,"%f\n",crossings[i]);

  }

  fclose(out);

 } else printf("Failed to open output file\n");


 sprintf(filename,"%s.samples.chan",src_file);
 printf("Writing sample data (%s)\n",filename);

 if (out = fopen(filename,"w")){

  for (i = 0; i < snd_info->samples; i++){

   fprintf(out,"%f\n",buffer[i]);

  }

  fclose(out);

 } else printf("Failed to open output file\n");

 free(buffer);
 free(lpc);
 free(indices);
 free(targets);
 free(tmp_target);

 sf_close(snd);

 printf("Closing error log\n");
 fclose(errorlog);

 return 0;

}



/*********************************************************************

Window: apply a window function to an array of samples

*********************************************************************/

void Window(double * buffer, int N, double a){

/*
 Hamming window: a = 0.54
 Hann(ing) window: a = 0.5;
*/

 int i = 0;

 for (i = 0; i < N; i++)

  buffer[i] *= ( a - ( 1 - a ) * cos( ( 2.0 * PI * (double) i ) / ( N - 1 ) ) );

}

/*********************************************************************

durbin: drive the main recursion (iteration) of Durbin's algorithm

*********************************************************************/

double durbin(double * a, double * window, int N, int p){

 double * k     = (double *) malloc((p+1) * sizeof(double));
 double * old_a = (double *) malloc((p+1) * sizeof(double));
 double * new_a = (double *) malloc((p+1) * sizeof(double));

 int j = 0, m = 0;

 double E = 0.0;

/* printf("durbin: started (N: %d, p: %d)\n", N, p); */

/* Begin */

 E = R(window, N, 0);
 k[1] = R(window, N, 1)/E;
 old_a[1] = k[1];

// printf("E(0): %f, a[1]: %f, k[1]: %f\n",E,old_a[1],k[1]);

 E *= (1 - k[1] * k[1]);

// printf("E(1): %f\n",E);

 for (j = 2; j <= p; j++){

  k[j] = parcor(window,N,j,E,old_a);

  new_a[j] = k[j];

  for (m = 1; m < j; m++) /* Determine all a's for this iteration */
   new_a[m] = old_a[m] - k[j] * old_a[j-m];

  E *= (1 - k[j] * k[j]);

//   printf("E(%d): %f\n",j,E);

  for (m = 1; m <= j; m++) old_a[m] = new_a[m];

 }

 for (m = 1; m <= p; m++) a[m] = new_a[m];


/* End */

 free(old_a);
 free(new_a);
 free(k);

 return E;

}


/*********************************************************************

parcor: determine the partial correlation coefficients for durbin

*********************************************************************/

double parcor(double * window, int N, int i, double E, double * a){

 double k = 0.0;

 int j = 0;

/* printf("parcor: started (N: %d, i: %d, E: %.4f)\n", N, i, E); */

 for (j = 1; j <= i - 1; j++)
  k += a[j] * R(window, N, i - j);

 return (R(window, N, i) - k) / E;

}


/*********************************************************************

R: determine the auto-correlation for durbin

*********************************************************************/

double R(double * window, int N, int k){

 int i;
 double r = 0.0;

/* printf("R: started (N: %d, k: %d)\n", N, k); */

 for (i = 0; i <= N - k - 1; i++)
  r += window[i] * window[i+k];

 return r;

}


/*********************************************************************

preEmphasis: apply pre-emphasis to an array of samples

*********************************************************************/

double * preEmphasis(double ** samples, int N, double a1){

 int i;
 double * newSamples = NULL;

 if (a1 != 0.0){

 printf("Applying pre-emphasis (%f)\n", a1);

  newSamples = (double *) malloc(sizeof(double) * N);

  newSamples[0] = (*samples)[0];

  for (i = 1; i < N; i++)
   newSamples[i] = (*samples)[i] - a1 * (*samples)[i-1];

  free(*samples);
  *samples = newSamples;

  return newSamples;

 } else return *samples;

}



/*********************************************************************

match: choose a template matching algorithm

*********************************************************************/

int * match(double * lpc, int p, double * distance, int intervals, double * targets, int n, int type){

/* Note p is same for targets & LPCs */

 int * indices = (int *) malloc(sizeof(int) * intervals);
 int i;

/* Normalise targets */

// normaliseTargets(targets, n, p);
 
// MATCH_EUCLIDEAN	1
// MATCH_SPECTRUM	2
// MATCH_MULTIPLY	3

// Use some template matching algorithm to find closest template */

 switch (type){

  case MATCH_EUCLIDEAN : // Find match with lowest Euclidean distance

   for (i = 0; i < intervals; i++)
    indices[i] = 1 + euclideanDistance(lpc + i * (p+1), targets, p, n, distance, i);

   break;

  case MATCH_MULTIPLY : // Half-arsed attempt at Ian's weird method

   for (i = 0; i < intervals; i++)
    indices[i] = 1 + multiply(lpc + i * (p+1), targets, p, n);

   break;

  case MATCH_SPECTRUM : // Euclidean distance between interval spectra

   spectra(indices, distance, intervals, lpc, targets, p, n); // loop in function

   break;

 } // switch

 return indices;

}


/*********************************************************************

spectra: Match the spectral intervals (C++)

*********************************************************************/

void spectra(int * indices, double * distance, int intervals, double * lpc, double * targets, int p, int n){

// Note: implemented in C++, unlike the rest of this program
// Spectrum has '-' overloaded to give Euclidean distance

 int i, j;
 double min = -1.0, thisDistance = 0.0;

// Now loop over all intervals and find closest match for each

 for (i = 0; i < intervals; i++){

  Spectrum current(lpc + i * (p+1) + 1, gain[i], p, DFT_SAMPLES);
  indices[i] = 0;
  min = -1.0, thisDistance = 0.0;
 
  for (j = 0; j < n; j++){

   Spectrum tmp(targets + j * (p+1) + 1, p, DFT_SAMPLES);
   thisDistance = current - tmp;

   if (thisDistance < min || min == -1.0){

    min = thisDistance;
    indices[i] = j + 1;
//    printf("Best match: %d -> %d (%f)\n",i,j+1,thisDistance);

   } // if

  } // j

  distance[i] = min;

 } // i

}



/*********************************************************************

euclideanDistance: calculate the Euclidian distance

*********************************************************************/

int euclideanDistance(double * lpc, double * targets, int p, int n, double * distance, int l){

 int i, j;
 int index = -1;
 double min = -1.0, sum;
 

 for (i = 0; i < n; i++){

  sum = 0.0;

  for (j = 1; j <= p; j++)
   sum += ( lpc[j] - targets[i*(p+1) + j] ) * ( lpc[j] - targets[i*(p+1) + j] );

  sum = sqrt(sum);

  if (sum < min || min == -1.0){
   min = sum;
   index = i;
  }

 } /* i */


//  Mahalanobis distance
//
// for (i = 0; i < n; i++){
//
//  sum = 0.0;
//
//  for (j = 1; j <= p; j++)
//   sum += pow(( lpc[j] - targets[i*(p+1) + j] ), lpc[j]) * ( lpc[j] - targets[i*(p+1) + j] );
//
//  sum = sqrt(sum);
//
//  if (sum < min || min == -1.0){
//   min = sum;
//   index = i;
//  }
//
// } /* i */

 distance[l] = min;

 return index; 

}

/*********************************************************************

multiply: experimental matching algorithm suggested by Ian Stephenson

*********************************************************************/

int multiply(double * lpc, double * targets, int p, int n){

 /* Assume normalised targets */

 double * tmpLPC = NULL;
 double sum = 0.0;
 double min = -1.0;
 
 int index = -1;
 int i, j; 
 
 tmpLPC = (double *) malloc(sizeof(double) * (p+1));

 /* Normalise LPCs */

 for (i = 1; i <= p; i++){
  tmpLPC[i] = lpc[i];
  sum += tmpLPC[i];
 }

 for (i = 1; i <= p; i++) tmpLPC[i] /= absd(sum);

 for (i = 0; i < n; i++){

  sum = 0.0;

  for (j = 1; j <= p; j++)
   sum += tmpLPC[j] * targets[i * (p+1) + j];
   
  if (sum < min || i == 0){
    min = sum;
    index = i;
  }

 } /* i */

 free(tmpLPC);
 return index;

}


/*********************************************************************

normaliseTargets: used by multiply, above

*********************************************************************/

void normaliseTargets(double * targets, int n, int p){

 int i, j;
 double sum = 0.0;

 printf("Normalising targets\n");

 for (i = 0; i < n; i++){

  sum = 0.0;
  
  for (j = 1; j <= p; j++) sum += targets[i*(p+1)+j];
  for (j = 1; j <= p; j++) targets[i*(p+1)+j] /= absd(sum);

 } /* i */
/*
 printf("Normalised targets:\n");
 for (i = 0; i < n*(p+1); i++) printf("%.3f\n",targets[i]);
*/
}


/*********************************************************************

absd: return the absolute value of a double

*********************************************************************/

double absd(double d){ return (d < 0.0) ? -1.0 * d : d ; }


/*********************************************************************

readTarget: extract average LPC's from a template sound file

*********************************************************************/

int readTarget(double ** targets, char * filename, double emphasis){

 SNDFILE * snd = NULL;
 SF_INFO * snd_info = (SF_INFO *) malloc(sizeof(SF_INFO));

 double * buffer = NULL;
 double * window = NULL;
 double * lpc = NULL;
 double * tmp_target = (double *) malloc(sizeof(double) * (order+1));

 double sum = 0.0;

 int Is = 0;
 int Ns = 0;
 int i = 0, j = 0;

 printf("Reading target \"%s\" (n: %d, order: %d)\n",filename, n_targets, order);

 if ((snd = sf_open_read(filename,snd_info)) != NULL){

  printf("Target file opened\n");

  buffer = (double *) malloc(sizeof(double)*snd_info->channels*snd_info->samples);
  sf_readf_double(snd,buffer,snd_info->samples,1);

  /* Process */

#ifndef NO_PRE_EMPHASIS
  preEmphasis(&buffer, snd_info->samples, emphasis); 
#endif
  /* LP for entire wave */

  Is = (int) ((double) snd_info->samplerate * I );	/* Length of window interval (samples) */
  Ns = (int) ((double) snd_info->samplerate * Nms );	/* Length of window (samples) */

  lpc = (double *) malloc(sizeof(double) * (order+1) * snd_info->samples / Is);
  window = (double *) malloc(sizeof(double) * Ns);

  for (i = 0; i < snd_info->samples / Is; i++){

   memcpy(window, buffer+(i*Is), Ns);
   Window(window, Ns, WINDOW_HAMMING);
   durbin(lpc+i*(order+1), window, Ns, order);

  }

  /* Average LPC's (no normalisation yet) and stick em in the targets array */
  /* Remember order 0 is unused in lpc[] */

  tmp_target[0] = 0.0;

  for (i = 1; i <= order; i++){

   sum = 0.0;

   for (j = 0; j < snd_info->samples / Is; j++)
    sum += lpc[j*(order+1)+i];

   sum /= snd_info->samples / Is;

   tmp_target[i] = sum;

   printf("\ta%d = %f\n",i,tmp_target[i]);

  }

  /* clear up */

  free(window);
  free(buffer);
  free(lpc);
  free(snd_info);
  sf_close(snd);

  n_targets++;

  *targets = (double *) realloc(*targets,n_targets*sizeof(double)*(order+1));
  memcpy(*targets+(n_targets-1)*(order+1),tmp_target,(order+1)*sizeof(double));

  free(tmp_target);

  fprintf(errorlog,"Target index %d read from file \"%s\"\n",n_targets,filename);

  return 0; /* Success, whoohoo */

 } else {

  free(snd_info);

  fprintf(errorlog,"Failed to read target file \"%s\"\n",filename);

  return -1; /* Failure, boooo */

 }

}


/*********************************************************************

Threshold: zero all target indices for less-than-threshold samples
I.e. set the blend shape to neutral when no sound

*********************************************************************/

void threshold(double * buffer, int * indices, int samples, int Is, double threshold){

/* If abs(signal) < threshold, blend index = 0 */
/* samples should be normalised */

 int i, j, silence;

 for (i = 0; i < samples/Is; i++){

  silence = 1;

  for (j = 0; j < Is; j++){

   if (absd(buffer[i*Is+j]) > threshold){
    silence = 0;
    break;
   }

  }

  if (silence) indices[i] = 0;

 }


}


/*********************************************************************

pitchEstimate: pitch estimate by auto-correlation method

*********************************************************************/

double pitchEstimate(double * correlation, int maxK, double triggerThreshold, double resetThreshold){

 int i = 0;
 int mode = PITCH_RESET; /* wait for normalised signal to drop below reset */
 int intervals = 1;
 int * indices = (int *) malloc(sizeof(int));

 double average = -1.0; 

 indices[0] = 0;

 for (i = 0; i < maxK; i++){

  if (mode == PITCH_RESET && correlation[i] < resetThreshold){

   mode = PITCH_TRIGGER;
   continue;

  }

  if (mode == PITCH_TRIGGER && correlation[i] > triggerThreshold){

   /* Trigger Event */

/*   printf("pitchEstimate: found trigger at %d samples (%f)\n",i,correlation[i]); */

   indices = (int *) realloc(indices, ++intervals * sizeof(int));
   indices[intervals-1] = i;

   mode = PITCH_RESET;
   continue;

  }

 } /* i */

 if (intervals > 1){
  average = 0;
  for (i = 1; i < intervals; i++)
   average += (double) indices[i] - (double) indices[i-1];
  average /= (double) (intervals - 1);
 }

 free(indices);

 return average;

}


/*********************************************************************

normaliseCovariance: normalise the auto-covariance, i.e return the
auto-correlation

*********************************************************************/

double * normaliseCovariance(double * covariance, int k){

 double * out = (double *) malloc((k+1) * sizeof(double));
 int i = 0;

 if (covariance[0]){

 if (out){

  for (i = k; i >= 0; i--) out[i] = covariance[i] / covariance[0];

 } /* if */

 } else printf("C(0) is zero, can't normalise\n");
 return out;

}

/*********************************************************************

autoCovariance: calculate the auto-covariance of a window

*********************************************************************/

double * autoCovariance(double * samples, int window, int maxK, int offset){

 double * out = (double *) malloc((maxK+1) * sizeof(double));
 int i = 0, k = 0;
 double * buffer = (double *) malloc(sizeof(double) * (window+maxK));

 if (out){

  for (i = 0; i < window + maxK; i++)
   buffer[i] = samples[offset+i];

//  Window(buffer,window,WINDOW_HAMMING);

  for (k = 0; k <= maxK; k++){

   out[k] = 0.0;

   for (i = 0; i <= window-1; i++){

    out[k] += buffer[i] * buffer[i+k];

   } /* i */

   out[k] /= window;

  } /* k */
 
 } /* if */

 free(buffer);
 return out;

}

/*********************************************************************

loadConfig: read config data from file, if it exists

*********************************************************************/

int loadConfig(double ** targets, int * p, double * thresh, double * trigger, double * release, double * emphasis){

 char buffer[256] = "";
 FILE * config;
 int lines = 0;
 int returned = 0;
 int data_lines = 0;

 if ((config = fopen(CONFIG_FILE,"r")) == NULL){

  printf("Failed to read config file\n");
  return -1;

 } else {

  printf("Reading config file \"%s\"\n",CONFIG_FILE);

  while((returned = readLine(config, buffer, 256)) != -1){

   if (returned > 0){

    lines++;

    switch (lines){

     case 1: // lines before first target file
      data_lines = atoi(buffer);
      break;

     case 2: // Predictor order
      if (lines <= data_lines){
       *p = atoi(buffer);
       printf("\torder:\t%d\n",*p);
      }
      break;

     case 3: // Threshold
      if (lines <= data_lines){
       *thresh = atof(buffer);
       printf("\tthreshold:\t%f\n",*thresh);
      }
      break;

     case 4: // Pitch Trigger Level 
      if (lines <= data_lines){
       *trigger = atof(buffer);
       printf("\ttrigger:\t%f\n",*trigger);
      }
      break;

     case 5: // Pitch Release Level 
      if (lines <= data_lines){
       *release = atof(buffer);
       printf("\trelease:\t%f\n",*release);
      }
      break;

     case 6: // Use pre-emphasis
      if (lines <= data_lines){
       *emphasis = atof(buffer);
       printf("\temphasis:\t%f\n",*emphasis);
      }
      break;

    }

    if (lines > data_lines && data_lines > 0) // targets
     readTarget(targets, buffer, *emphasis);

   } // if

  } // while

  return lines;

 } // if

} // loadConfig




/*********************************************************************

readLine: read a line from the config file

*********************************************************************/

int readLine(FILE * filePtr, char * buffer, int bufsize){

 int i = 0;
 char tmp = fgetc(filePtr);

  while (!feof(filePtr) &&
         i < bufsize &&
         tmp != 10){

   /* printf("%d: %c (%d)\n",i,tmp,(int)tmp); */
 
   buffer[i++] = tmp;

   tmp = fgetc(filePtr);
  }

 /* Null-terminate the string */

 buffer[i] = '\0';

 if (feof(filePtr)) return -1;
 else return i;

}


/*********************************************************************

sign: return the sign of a double

*********************************************************************/

int sign(double x){ return (x >= 0.0) ? 1 : -1; }



/*********************************************************************

ZeroCrossing: count zero-crossings

*********************************************************************/

double * ZeroCrossing(double * samples, int nSamples, int window){

 double * crossings = (double *) malloc(sizeof(double) * nSamples);
 double * result = (double *) malloc(sizeof(double) * nSamples);
 double * buffer = (double *) malloc(sizeof(double) * window);
 double sum = 0.0;

 int i = 0, j = 0;

 if (crossings){

  for (i = 1; i < nSamples; i++)
   crossings[i] = (double) abs( sign(samples[i]) - sign(samples[i-1]) ) / 2.0;

  printf("Counting zero-crossings (%d)\n", window);

  for (i = 0; i < nSamples; i++){

   if (i < window/2) result[i] = 0.0;
   else if (i > nSamples-window/2) result[i] = 0.0;
   else {

    sum = 0;

    for (j = 0; j < window; j++) buffer[j] = crossings[i-window/2+j];
    Window(buffer,window,WINDOW_HAMMING);
    for (j = 0; j < window; j++) sum += buffer[j];
    result[i] = sum / (double) window;

   } /* if */

  } /* i */

 } else printf("ZeroCrossing: Failed memory allocation\n");

 free(buffer);
 free(crossings);

 return result;

}



/*********************************************************************

average: return the average of a set of samples. Originally used in
lieu of a gain function

*********************************************************************/

double average(double * data, int n){

/* Return the average of a set of samples */

 int i;
 double sum = 0;

 for (i = 0; i < n; i++) sum += data[i];

 return sum/(double)n;

}



/*********************************************************************

G: calculate the gain of a bunch of samples

*********************************************************************/


double G(double * lpc, int p, double * data, int n){

/* Evaluate the gain of a set of samples  */

 int i;
 double sum = 0;

 for (i = 1; i <= p; i++) sum += data[i] * R(data,n,i);

 return sqrt(R(data,n,0) - sum);

}

